import time
from typing import Optional, Generic


from causy.interfaces import (
    LogicStepInterface,
    BaseGraphInterface,
    GraphModelInterface,
    ExitConditionInterface,
    LogicStepInterfaceType,
)

from causy.models import ActionHistoryStep


class Loop(LogicStepInterface[LogicStepInterfaceType], Generic[LogicStepInterfaceType]):
    """
    A loop which executes a list of pipeline_steps until the exit_condition is met.
    """

    exit_condition: Optional[ExitConditionInterface] = None

    def execute(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        """
        Executes the loop til self.exit_condition is met
        :param graph:
        :param graph_model_instance_:
        :return:
        """
        n = 0
        steps = []
        loop_started = time.time()
        actions_taken = None
        while not self.exit_condition(
            graph=graph,
            graph_model_instance_=graph_model_instance_,
            actions_taken=actions_taken,
            iteration=n,
        ):
            actions_taken = []
            for pipeline_step in self.pipeline_steps:
                started = time.time()
                (
                    actions_taken,
                    all_proposed_actions,
                ) = graph_model_instance_.execute_pipeline_step(pipeline_step)
                steps.append(
                    ActionHistoryStep(
                        name=pipeline_step.name,
                        actions=actions_taken,
                        all_proposed_actions=all_proposed_actions,
                        duration=time.time() - started,
                    )
                )
                actions_taken.extend(actions_taken)
            n += 1
        return ActionHistoryStep(
            name=self.name,
            steps=steps,
            duration=time.time() - loop_started,
        )


class ApplyActionsTogether(
    LogicStepInterface[LogicStepInterfaceType], Generic[LogicStepInterfaceType]
):
    """
    A logic step which collects all actions and only takes them at the end of the pipeline
    """

    def execute(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        """
        Executes the loop til self.exit_condition is met
        :param graph:
        :param graph_model_instance_:
        :return:
        """
        actions = []
        steps = []
        loop_started = time.time()
        for pipeline_step in self.pipeline_steps:
            started = time.time()
            (
                actions_taken,
                all_proposed_actions,
            ) = graph_model_instance_.execute_pipeline_step(
                pipeline_step, apply_to_graph=False
            )
            steps.append(
                ActionHistoryStep(
                    name=pipeline_step.name,
                    actions=actions_taken,
                    all_proposed_actions=all_proposed_actions,
                    duration=time.time() - started,
                )
            )
            actions.extend(actions_taken)

        graph_model_instance_._take_action(actions)

        return ActionHistoryStep(
            name=self.name,
            steps=steps,
            duration=time.time() - loop_started,
        )



import enum
from datetime import datetime
from typing import Optional, Dict, List, Union, Any

from pydantic import BaseModel, computed_field, Field

from causy.graph_utils import serialize_module_name, hash_dictionary
from causy.interfaces import (
    AS_MANY_AS_FIELDS,
    NodeInterface,
    PipelineStepInterfaceType,
    LogicStepInterface,
    EdgeTypeInterface,
    ExtensionType,
    EdgeInterface,
    TestResultInterface,
    ComparisonSettingsInterface,
    ExtensionInterface,
    EdgeTypeInterfaceType,
    GraphUpdateHook,
)
from causy.variables import IntegerParameter, VariableInterfaceType


class ComparisonSettings(ComparisonSettingsInterface):
    min: IntegerParameter = 2
    max: IntegerParameter = AS_MANY_AS_FIELDS

    @computed_field
    @property
    def name(self) -> str:
        return serialize_module_name(self)


class TestResultAction(str, enum.Enum):
    """
    Actions that can be taken on the graph. These actions are used to keep track of the history of the graph.
    """

    UPDATE_EDGE = "UPDATE_EDGE"
    UPDATE_EDGE_DIRECTED = "UPDATE_EDGE_DIRECTED"

    UPDATE_EDGE_TYPE = "UPDATE_EDGE_TYPE"
    UPDATE_EDGE_TYPE_DIRECTED = "UPDATE_EDGE_TYPE_DIRECTED"

    REMOVE_EDGE_UNDIRECTED = "REMOVE_EDGE_UNDIRECTED"
    REMOVE_EDGE_DIRECTED = "REMOVE_EDGE_DIRECTED"

    RESTORE_EDGE = "RESTORE_EDGE"
    RESTORE_EDGE_DIRECTED = "RESTORE_EDGE_DIRECTED"

    DO_NOTHING = "DO_NOTHING"


class TestResult(TestResultInterface):
    u: NodeInterface
    v: NodeInterface
    action: TestResultAction
    data: Optional[Dict] = None


class AlgorithmReferenceType(str, enum.Enum):
    FILE = "file"
    NAME = "name"
    PYTHON_MODULE = "python_module"


class AlgorithmReference(BaseModel):
    reference: str
    type: AlgorithmReferenceType


class Algorithm(BaseModel):
    name: str
    pipeline_steps: List[Union[PipelineStepInterfaceType, LogicStepInterface]]
    pipeline_steps: List[Union[PipelineStepInterfaceType, LogicStepInterface]]
    edge_types: List[EdgeTypeInterfaceType]
    extensions: Optional[List[ExtensionType]] = None
    variables: Optional[List[Union[VariableInterfaceType]]] = None
    pre_graph_update_hooks: Optional[List[GraphUpdateHook]] = None
    post_graph_update_hooks: Optional[List[GraphUpdateHook]] = None

    def hash(self) -> str:
        return hash_dictionary(self.model_dump())


class ActionHistoryStep(BaseModel):
    name: str
    duration: Optional[float] = None  # seconds
    actions: Optional[List[TestResult]] = []
    all_proposed_actions: Optional[
        List[TestResult]
    ] = []  # all actions and not only the ones that were executed
    steps: Optional[List["ActionHistoryStep"]] = []


class Result(BaseModel):
    algorithm: AlgorithmReference
    created_at: datetime = Field(default_factory=datetime.now)
    nodes: Dict[str, NodeInterface]
    edges: List[EdgeInterface]
    action_history: List[ActionHistoryStep]
    variables: Optional[Dict[str, Any]] = None
    data_loader_hash: Optional[str] = None
    algorithm_hash: Optional[str] = None
    variables_hash: Optional[str] = None



import enum
import hashlib
import importlib
import json
from abc import ABC, abstractmethod
from typing import Iterator, Dict, Union, Any, Optional

from pydantic import BaseModel

from causy.graph_utils import hash_dictionary


class DataLoaderType(str, enum.Enum):
    DYNAMIC = "dynamic"  # python function which yields data
    JSON = "json"
    JSONL = "jsonl"


class DataLoaderReference(BaseModel):
    """represents a single data loader
    :param type: the type of dataloader
    :param reference: path to either the python class which can be executed to load the data or the data source file itself
    """

    type: DataLoaderType
    reference: str
    options: Optional[Dict[str, Any]] = None


class AbstractDataLoader(ABC):
    @abstractmethod
    def __init__(self, reference: str, options: Optional[Dict[str, Any]] = None):
        pass

    reference: str
    options: Optional[Dict[str, Any]] = None

    @abstractmethod
    def load(self) -> Iterator[Dict[str, Union[float, int, str]]]:
        """
        loads the data from the source and returns it as an iterator
        :return:
        """
        pass

    @abstractmethod
    def hash(self) -> str:
        """
        returns a hash of the data that is loaded
        :return:
        """
        pass

    def _hash_options(self):
        return hash_dictionary(self.options)


class FileDataLoader(AbstractDataLoader, ABC):
    """
    A data loader which loads data from a file reference (e.g. json, csv, etc.)
    """

    def __init__(self, reference: str, options: Optional[Dict[str, Any]] = None):
        self.reference = reference
        self.options = options

    reference: str

    def hash(self) -> str:
        with open(self.reference, "rb") as f:
            return (
                f'{hashlib.file_digest(f, "sha256").hexdigest()}_{self._hash_options()}'
            )


class JSONDataLoader(FileDataLoader):
    """
    A data loader which loads data from a json file
    """

    def load(self) -> Iterator[Dict[str, Union[float, int, str]]]:
        with open(self.reference, "r") as f:
            data = json.loads(f.read())
            if isinstance(data, list):
                for item in data:
                    yield item
            elif isinstance(data, dict):
                yield {"_dict": data}
                return
            else:
                raise ValueError(
                    f"Invalid JSON format. Data in {self.reference} is of type {type(data)}."
                )


class JSONLDataLoader(FileDataLoader):
    """
    A data loader which loads data from a jsonl file
    """

    def load(self) -> Iterator[Dict[str, Union[float, int, str]]]:
        with open(self.reference, "r") as f:
            for line in f:
                yield json.loads(line)


class DynamicDataLoader(AbstractDataLoader):
    """
    A data loader which loads another data loader dynamically based on the reference
    """

    def __init__(self, reference: str, options: Optional[Dict[str, Any]] = None):
        self.reference = reference
        self.data_loader = self._load_data_loader()
        self.options = options

    reference: str
    data_loader: AbstractDataLoader

    def _load_data_loader(self) -> AbstractDataLoader:
        module = importlib.import_module(self.reference)
        # todo: should the cls be referenced here?
        return module.DataLoader(**self.options)

    def load(self) -> Iterator[Dict[str, Union[float, int, str]]]:
        return self.data_loader.load()


DATA_LOADERS = {
    DataLoaderType.JSON: JSONDataLoader,
    DataLoaderType.JSONL: JSONLDataLoader,
    DataLoaderType.DYNAMIC: DynamicDataLoader,
}


def load_data_loader(reference: DataLoaderReference) -> AbstractDataLoader:
    """
    loads the data loader based on the reference
    :param reference: a data loader reference
    :return:
    """

    return DATA_LOADERS[reference.type](reference.reference, reference.options)



import json
import logging

import typer

from causy.data_loader import JSONDataLoader
from causy.graph_model import graph_model_factory
from causy.models import (
    Result,
    AlgorithmReferenceType,
)
from causy.serialization import (
    serialize_algorithm,
    load_algorithm_from_specification,
    CausyJSONEncoder,
    load_json,
)
from causy.graph_utils import (
    retrieve_edges,
)
from causy.ui.cli import ui as ui_app
from causy.workspaces.cli import workspace_app as workspaces_app
from causy.causal_discovery import AVAILABLE_ALGORITHMS

app = typer.Typer()

app.add_typer(workspaces_app, name="workspace")
app.command(name="ui", help="run causy ui")(ui_app)


@app.command()
def eject(algorithm: str, output_file: str):
    logging.warning(
        f"Ejecting pipelines outside of workspace context is deprecated. Please use workspaces instead."
    )

    typer.echo(f"💾 Loading algorithm {algorithm}")
    model = AVAILABLE_ALGORITHMS[algorithm]()
    result = serialize_algorithm(model, algorithm_name=algorithm)
    typer.echo(f"💾 Saving algorithm {algorithm} to {output_file}")
    with open(output_file, "w") as file:
        file.write(json.dumps(result, indent=4))


@app.command()
def execute(
    data_file: str,
    pipeline: str = None,
    algorithm: str = None,
    output_file: str = None,
    log_level: str = "ERROR",
):
    logging.warning(
        f"Executing outside of workspaces is deprecated and will be removed in future versions. Please use workspaces instead."
    )
    logging.basicConfig(level=log_level)
    if pipeline:
        typer.echo(f"💾 Loading pipeline from {pipeline}")
        model_dict = load_json(pipeline)
        algorithm = load_algorithm_from_specification(model_dict)
        model = graph_model_factory(algorithm=algorithm)()
        algorithm_reference = {
            "type": AlgorithmReferenceType.FILE,
            "reference": pipeline,  # TODO: how to reference pipeline in a way that it can be loaded?
        }
    elif algorithm:
        typer.echo(f"💾 Creating pipeline from algorithm {algorithm}")
        model = AVAILABLE_ALGORITHMS[algorithm]()
        algorithm_reference = {
            "type": AlgorithmReferenceType.NAME,
            "reference": algorithm,
        }

    else:
        raise ValueError("Either pipeline_file or algorithm must be specified")

    dl = JSONDataLoader(data_file)
    # initialize from json
    model.create_graph_from_data(dl)

    # TODO: I should become a configurable skeleton builder
    model.create_all_possible_edges()

    typer.echo("🕵🏻‍♀  Executing pipeline steps...")
    model.execute_pipeline_steps()
    edges = []
    for edge in retrieve_edges(model.graph):
        print(
            f"{model.graph.nodes[edge[0]].name} -> {model.graph.nodes[edge[1]].name}: {model.graph.edges[edge[0]][edge[1]]}"
        )

    result = Result(
        algorithm=algorithm_reference,
        action_history=model.graph.graph.action_history,
        edges=model.graph.retrieve_edges(),
        nodes=model.graph.nodes,
    )

    if output_file:
        typer.echo(f"💾 Saving graph actions to {output_file}")
        with open(output_file, "w") as file:
            file.write(json.dumps(result.model_dump(), cls=CausyJSONEncoder, indent=4))


if __name__ == "__main__":
    app()



import logging
from typing import Tuple, List, Generic, Optional

from causy.interfaces import (
    PipelineStepInterface,
    BaseGraphInterface,
    PipelineStepInterfaceType,
)
from causy.models import TestResultAction, TestResult
from causy.variables import StringVariable, IntegerVariable, FloatVariable, BoolVariable

logger = logging.getLogger(__name__)


class PlaceholderTest(
    PipelineStepInterface[PipelineStepInterfaceType], Generic[PipelineStepInterfaceType]
):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if "placeholder_str" in kwargs:
            self.placeholder_str = kwargs["placeholder_str"]

        if "placeholder_int" in kwargs:
            self.placeholder_int = kwargs["placeholder_int"]

        if "placeholder_float" in kwargs:
            self.placeholder_float = kwargs["placeholder_float"]

        if "placeholder_bool" in kwargs:
            self.placeholder_bool = kwargs["placeholder_bool"]

    placeholder_str: Optional[StringVariable] = "placeholder"
    placeholder_int: Optional[IntegerVariable] = 1
    placeholder_float: Optional[FloatVariable] = 1.0
    placeholder_bool: Optional[BoolVariable] = True

    def process(
        self, nodes: Tuple[str], graph: BaseGraphInterface
    ) -> List[TestResult] | TestResult:
        """
        Placeholder test for testing purposes
        :param nodes:
        :param graph:
        :return:
        """
        logger.debug(f"PlaceholderTest {nodes}")
        return TestResult(u=None, v=None, action=TestResultAction.DO_NOTHING, data={})



import enum
from typing import Optional, List, Generic

from pydantic import BaseModel

from causy.interfaces import ExtensionInterface, ExtensionType


class EdgeUIConfig(BaseModel):
    color: Optional[str] = None
    width: Optional[int] = None
    style: Optional[str] = None
    marker_start: Optional[str] = None
    marker_end: Optional[str] = None
    label_field: Optional[str] = None
    animated: Optional[bool] = False
    label: Optional[str] = None


class ConditionalEdgeUIConfigComparison(str, enum.Enum):
    EQUAL = "EQUAL"
    GREATER = "GREATER"
    LESS = "LESS"
    GREATER_EQUAL = "GREATER_EQUAL"
    LESS_EQUAL = "LESS_EQUAL"
    NOT_EQUAL = "NOT_EQUAL"


class ConditionalEdgeUIConfig(BaseModel):
    ui_config: Optional[EdgeUIConfig] = None
    condition_field: str = "coefficient"
    condition_value: float = 0.5
    condition_comparison: ConditionalEdgeUIConfigComparison = (
        ConditionalEdgeUIConfigComparison.GREATER_EQUAL
    )


class EdgeTypeConfig(BaseModel):
    edge_type: str
    default_ui_config: Optional[EdgeUIConfig] = None
    conditional_ui_configs: Optional[List[ConditionalEdgeUIConfig]] = None


class GraphUIExtension(ExtensionInterface[ExtensionType], Generic[ExtensionType]):
    edges: List[EdgeTypeConfig]



from typing import Tuple, Optional, Generic

import torch

from causy.generators import AllCombinationsGenerator
from causy.interfaces import (
    PipelineStepInterface,
    BaseGraphInterface,
    GeneratorInterface,
    PipelineStepInterfaceType,
)
from causy.models import ComparisonSettings, TestResultAction, TestResult


class CalculatePearsonCorrelations(
    PipelineStepInterface[PipelineStepInterfaceType], Generic[PipelineStepInterfaceType]
):
    generator: Optional[GeneratorInterface] = AllCombinationsGenerator(
        comparison_settings=ComparisonSettings(min=2, max=2)
    )
    chunk_size_parallel_processing: int = 1
    parallel: bool = False

    def process(self, nodes: Tuple[str], graph: BaseGraphInterface) -> TestResult:
        """
        Calculate the correlation between each pair of nodes and store it to the respective edge.
        :param nodes: list of nodes
        :return: A TestResult with the action to take
        """
        x = graph.nodes[nodes[0]]
        y = graph.nodes[nodes[1]]
        edge_value = graph.edge_value(graph.nodes[nodes[0]], graph.nodes[nodes[1]])

        x_val = x.values
        y_val = y.values

        cov_xy = torch.mean((x_val - x_val.mean()) * (y_val - y_val.mean()))
        std_x = x_val.std(unbiased=False)
        std_y = y_val.std(unbiased=False)
        pearson_correlation = cov_xy / (std_x * std_y)

        correlation = pearson_correlation.item()

        # Clamp the correlation to -1 and 1 to avoid numerical errors
        if correlation < -1:
            correlation = -1
        elif correlation > 1:
            correlation = 1

        edge_value["correlation"] = correlation

        return TestResult(
            u=x,
            v=y,
            action=TestResultAction.UPDATE_EDGE,
            data=edge_value,
        )



import collections
from dataclasses import dataclass
from typing import List, Optional, Dict, Set, Tuple, Union, OrderedDict, Any
from uuid import uuid4
import logging

import torch
from pydantic import BaseModel, Field

from causy.edge_types import UndirectedEdge, DirectedEdge
from causy.interfaces import (
    BaseGraphInterface,
    NodeInterface,
    EdgeInterface,
    EdgeTypeInterface,
    MetadataType,
)
from causy.models import TestResultAction, TestResult, ActionHistoryStep
from causy.variables import VariableType

logger = logging.getLogger(__name__)


class Node(NodeInterface):
    """
    A node is a variable in the graph. It has a name, an id and values. The values are stored as a torch.Tensor.
    A node can also have metadata.
    """

    name: str
    id: str
    values: Optional[torch.Tensor] = None
    metadata: Optional[Dict[str, MetadataType]] = None

    def __hash__(self):
        return hash(self.id)


@dataclass
class Edge(EdgeInterface):
    """
    An edge is a connection between two nodes. It has a direction defined by the edge_type and metadata.
    A metadata example could be the p-value of a statistical test.
    An edge can be e.g. undirected, directed or bidirected. Which option is available/chosen depends on the algorithm and the current state of the graph.
    By default an edge always has an undirected and a directed edge_type.
    """

    u: NodeInterface
    v: NodeInterface
    edge_type: EdgeTypeInterface
    metadata: Optional[Dict[str, MetadataType]] = None
    deleted: Optional[bool] = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __hash__(self):
        return hash((self.u.id, self.v.id))


class GraphError(Exception):
    pass


class GraphBaseAccessMixin:
    def directed_edge_is_soft_deleted(
        self, u: Union[Node, str], v: Union[Node, str]
    ) -> bool:
        """
        Check if an edge is soft deleted
        :param u:
        :param v:
        :return:
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if u not in self.edges:
            return False
        if v not in self.edges[u]:
            return False
        return self.edges[u][v].deleted

    def parents_of_node(self, u: Union[Node, str]):
        """
        Return all parents of a node u
        :param u: node u
        :return: list of nodes (parents)
        """
        if isinstance(u, Node):
            u = u.id

        return [
            self.nodes[n]
            for n in self._reverse_edges[u].keys()
            if not self.directed_edge_is_soft_deleted(n, u)
        ]

    def edge_exists(self, u: Union[Node, str], v: Union[Node, str]) -> bool:
        """
        Check if any edge exists between u and v. Cases: u -> v, u <-> v, u <- v
        :param u: node u
        :param v: node v
        :return: True if any edge exists, False otherwise
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if u not in self.nodes:
            return False
        if v not in self.nodes:
            return False
        if (
            u in self.edges
            and v in self.edges[u]
            and not self.directed_edge_is_soft_deleted(u, v)
        ):
            return True
        if (
            v in self.edges
            and u in self.edges[v]
            and not self.directed_edge_is_soft_deleted(v, u)
        ):
            return True
        return False

    def directed_edge_exists(self, u: Union[Node, str], v: Union[Node, str]) -> bool:
        """
        Check if a directed edge exists between u and v. Cases: u -> v, u <-> v
        :param u: node u
        :param v: node v
        :return: True if a directed edge exists, False otherwise
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if u not in self.nodes:
            return False
        if v not in self.nodes:
            return False
        if u not in self.edges:
            return False
        if v not in self.edges[u]:
            return False

        if self.directed_edge_is_soft_deleted(u, v):
            return False

        return True

    def node_by_id(self, id_: str) -> Optional[Node]:
        """
        Retrieve a node by its id
        :param id_:
        :return:
        """
        return self.nodes.get(id_)

    def edge_value(self, u: Union[Node, str], v: Union[Node, str]) -> Optional[Dict]:
        """
        retrieve the value of an edge
        :param u:
        :param v:
        :return:
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if not self.edge_exists(u, v):
            return None
        return self.edges[u][v].metadata

    def edge_type(
        self, u: Union[Node, str], v: Union[Node, str]
    ) -> Optional[EdgeTypeInterface]:
        """
        retrieve the value of an edge
        :param u:
        :param v:
        :return:
        """
        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id
        if not self.edge_exists(u, v):
            return None
        return self.edges[u][v].edge_type

    def undirected_edge_exists(self, u: Union[Node, str], v: Union[Node, str]) -> bool:
        """
        Check if an undirected edge exists between u and v. Note: currently, an undirected edges is implemented just as
        a directed edge. However, they are two functions as they mean different things in different algorithms.
        Currently, this function is used in the PC algorithm, where an undirected edge is an edge which could not be
        oriented in any direction by orientation rules.
        Later, a cohersive naming scheme should be implemented.
        :param u: node u
        :param v: node v
        :return: True if an undirected edge exists, False otherwise
        """

        if self.directed_edge_exists(u, v) and self.directed_edge_exists(v, u):
            return True
        return False

    def get_siblings(self, v: Union[Node, str]) -> Set[Union[Node, str]]:
        """
        Get the set of nodes that are connected to the node v with an undirected edge.
        :param v: node v
        :return: A set of nodes that are connected to v with an undirected edge.
        """
        siblings = set()

        # Assuming self.nodes is a list or set of all nodes in the graph
        for node in self.nodes:
            if node != v and self.undirected_edge_exists(v, node):
                siblings.add(node)

        return siblings

    def retrieve_edge_history(
        self, u: Union[Node, str], v: Union[Node, str], action: TestResultAction = None
    ) -> List[TestResult]:
        """
        Retrieve the edge history
        :param u:
        :param v:
        :param action:
        :return:
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if action is None:
            return self.edge_history[(u, v)]

        if (u, v) not in self.edge_history:
            return []

        return [i for i in self.edge_history[(u, v)] if i.action == action]

    def directed_path_exists(self, u: Union[Node, str], v: Union[Node, str]) -> bool:
        """
        Check if a directed path from u to v exists
        :param u: node u
        :param v: node v
        :return: True if a directed path exists, False otherwise
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if self.directed_edge_exists(u, v):
            return True
        for w in self.edges[u]:
            if self.directed_path_exists(self.nodes[w], v):
                return True
        return False

    def edge_of_type_exists(
        self,
        u: Union[Node, str],
        v: Union[Node, str],
        edge_type: EdgeTypeInterface = DirectedEdge(),
    ) -> bool:
        """
        Check if an edge of a specific type exists between u and v.
        :param u: node u
        :param v: node v
        :param edge_type: the type of the edge to check for
        :return: True if an edge of this type exists, False otherwise
        """

        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        if not self.directed_edge_exists(u, v):
            return False

        if self.edges[u][v].edge_type != edge_type:
            return False

        return True

    def descendants_of_node(
        self, u: Union[Node, str], visited=None
    ) -> List[Union[Node, str]]:
        """
        Returns a list of all descendants of the given node in a directed graph, including the input node itself.
        :param u: The node or node ID to find descendants for.
        :return: A list of descendants of a node including the input node itself.
        """

        if visited is None:
            visited = set()

        if isinstance(u, Node):
            u = u.id

        # If this node has already been visited, return an empty list to avoid cycles
        if u in visited:
            return []

        visited.add(u)
        descendants = [self.node_by_id(u)]

        if u not in self.edges:
            return descendants

        for child in self.edges[u]:
            if self.directed_edge_exists(u, child) and not self.directed_edge_exists(
                child, u
            ):
                if child not in visited:
                    descendants.extend(self.descendants_of_node(child, visited))

        return list(set(descendants))

    def _is_a_collider_blocking(self, path, conditioning_set) -> bool:
        """
        Check if a path is blocked by a collider which is not in the conditioning set and has no descendants in the conditioning set.
        :return: Boolean indicating if the path is blocked
        """
        is_path_blocked = False
        for i in range(1, len(path) - 1):
            if self._is_collider(path[i - 1], path[i], path[i + 1]):
                # if the node is a collider, check if the node or any of its descendants are in the conditioning set
                is_path_blocked = True
                for descendant in self.descendants_of_node(path[i]):
                    if descendant in conditioning_set:
                        is_path_blocked = False
        return is_path_blocked

    def _is_a_non_collider_in_conditioning_set(self, path, conditioning_set) -> bool:
        """
        Check if a path is blocked by a non-collider which is in the conditioning set.
        :param path:
        :param conditioning_set:
        :return:
        """
        is_path_blocked = False
        for i in range(1, len(path) - 1):
            if path[i] in conditioning_set:
                # make sure that node is a noncollider
                if not (
                    self.edge_of_type_exists(path[i - 1].id, path[i].id, DirectedEdge())
                    and self.edge_of_type_exists(
                        path[i + 1].id, path[i].id, DirectedEdge()
                    )
                ):
                    # if the node is a non-collider and in the conditioning set, the path is blocked
                    is_path_blocked = True
        return is_path_blocked

    def are_nodes_d_separated_dag(
        self,
        u: Union[Node, str],
        v: Union[Node, str],
        conditioning_set: List[Union[Node, str]],
    ) -> bool:
        """
        Check if nodes u and v are d-separated given a conditioning set in a DAG. We check whether there is an open path, i.e. a path on which all colliders are in the conditioning set and all non-colliders are not in the conditioning set. If there is no open path, u and v are d-separated.

        :param u: First node
        :param v: Second node
        :param conditioning_set: Set of nodes to condition on
        :return: True if u and v are d-separated given conditioning_set, False otherwise
        """

        # Convert Node instances to their IDs
        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        # u and v may not be in the conditioning set, throw error
        if u in conditioning_set or v in conditioning_set:
            raise ValueError("Nodes u and v may not be in the conditioning set")

        # If there are no paths between u and v, they are d-separated
        if list(self.all_paths_on_underlying_undirected_graph(u, v)) == []:
            return True

        list_of_results_for_paths = []
        for path in self.all_paths_on_underlying_undirected_graph(u, v):
            # If the path only has two nodes, it cannot be blocked and is open. Therefore, u and v are not d-separated
            if len(path) == 2:
                return False

            is_path_blocked = False

            if self._is_a_collider_blocking(
                path, conditioning_set
            ) or self._is_a_non_collider_in_conditioning_set(path, conditioning_set):
                is_path_blocked = True

            list_of_results_for_paths.append(is_path_blocked)

        # if there is at least one open path, u and v are not d-separated
        if False in list_of_results_for_paths:
            return False
        return True

    def are_nodes_d_separated_cpdag(
        self,
        u: Union[Node, str],
        v: Union[Node, str],
        conditioning_set: List[Union[Node, str]],
    ) -> bool:
        """
        Check if nodes u and v are d-separated given a conditioning set in a CPDAG following Perkovic, 2020. (Identifying causal effects in maximally oriented partially directed acyclic graphs)

        :param u: First node
        :param v: Second node
        :param conditioning_set: Set of nodes to condition on
        :return: True if u and v are d-separated given conditioning_set, False otherwise
        """

        # Convert Node instances to their IDs
        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        # u and v may not be in the conditioning set, throw error
        if u in conditioning_set or v in conditioning_set:
            raise ValueError("Nodes u and v may not be in the conditioning set")

        # If there are no paths between u and v, they are d-separated
        if list(self.all_paths_on_underlying_undirected_graph(u, v)) == []:
            return True

        list_of_results_for_paths = []
        for path in self.all_paths_on_underlying_undirected_graph(u, v):
            # If the path only has two nodes, it cannot be blocked and is open. Therefore, u and v are not d-separated
            if len(path) == 2:
                return False
            is_path_blocked = False
            if self._is_path_of_definite_status(path):
                if self._is_definite_noncollider_in_conditioning_set(
                    path, conditioning_set
                ):
                    is_path_blocked = True
                if self._is_a_collider_blocking(path, conditioning_set):
                    is_path_blocked = True
            list_of_results_for_paths.append(is_path_blocked)
        if False in list_of_results_for_paths:
            return False
        return True

    def _is_collider(self, u: Node, v: Node, w: Node) -> bool:
        """
        Check if a node is a collider in a triple u -> v <- w
        :param u: Node u
        :param v: Node v
        :param w: Node w
        :return: True if the node is a collider, False otherwise
        """
        if self.edge_of_type_exists(u, v, DirectedEdge()) and self.edge_of_type_exists(
            w, v, DirectedEdge()
        ):
            return True
        return False

    def _is_definite_noncollider(self, u: Node, v: Node, w: Node) -> bool:
        """
        Check if a node is a definite non-collider in a triple u - v - w
        :param u: Node u
        :param v: Node v
        :param w: Node w
        :return: True if the node is a definite non-collider, False otherwise
        """
        # if there is an outgoing edge from the middle node, it is a definite non-collider
        if self.edge_of_type_exists(v, u, DirectedEdge()) or self.edge_of_type_exists(
            v, w, DirectedEdge()
        ):
            return True
        # if there is no edge between the outer nodes and it is not a collider, it is a definite non-collider
        if (not self.edge_exists(u, w)) and (not self._is_collider(u, v, w)):
            return True
        return False

    def _is_path_of_definite_status(self, path: List[Node]):
        """
        Check if a path is of definite status, i.e. if every node is either the start or end node, a collider or a definite non-collider (i.e. has an outgoing edge or is an unshielded triple on the path)
        :param path: The path to check
        :return: True if the path is of definite status, False otherwise
        """

        for i in range(1, len(path) - 1):
            # is node a collider?
            if self._is_collider(path[i - 1], path[i], path[i + 1]):
                continue
            # is node a definite non-collider because it has an outgoing edge?
            elif self._is_definite_noncollider(path[i - 1], path[i], path[i + 1]):
                continue
            return False
        return True

    def _is_definite_noncollider_in_conditioning_set(
        self, path, conditioning_set
    ) -> bool:
        """
        Check if a path is blocked by a definite non-collider which is in the conditioning set.
        :param path:
        :param conditioning_set:
        :return:
        """
        is_path_blocked = False
        for i in range(1, len(path) - 1):
            if path[i] in conditioning_set and self._is_definite_noncollider(
                path[i - 1], path[i], path[i + 1]
            ):
                is_path_blocked = True
        return is_path_blocked

    def all_paths_on_underlying_undirected_graph(
        self, u: Union[Node, str], v: Union[Node, str], visited=None, path=None
    ) -> List[List[Node]]:
        if isinstance(u, Node):
            u = u.id
        if isinstance(v, Node):
            v = v.id

        # Initialize visited and path lists during the first call
        if visited is None:
            visited = set()
        if path is None:
            path = []

        # Add the current node to the path and mark it as visited
        visited.add(u)
        path.append(u)

        # If the current node is the target node, yield the current path
        if u == v:
            yield list([self.node_by_id(node_id) for node_id in path])
        else:
            # Explore all possible next nodes by checking edge existence
            for next_node in self.nodes:
                if next_node not in visited and self.edge_exists(u, next_node):
                    yield from self.all_paths_on_underlying_undirected_graph(
                        next_node, v, visited, path
                    )

        # Backtrack: remove the current node from the path and visited set
        path.pop()
        visited.remove(u)

    def retrieve_edges(self) -> List[Edge]:
        """
        Retrieve all edges
        :return: all edges
        """
        edges = []
        for u in self.edges:
            for v in self.edges[u]:
                if not self.directed_edge_is_soft_deleted(self.nodes[u], self.nodes[v]):
                    edges.append(self.edges[u][v])
        return edges


class Graph(BaseModel, GraphBaseAccessMixin):
    nodes: OrderedDict[str, Node] = collections.OrderedDict({})
    edges: Dict[str, Dict[str, Edge]] = dict()
    _reverse_edges: Dict[str, Dict[str, Edge]] = dict()
    _deleted_edges: Dict[str, Dict[str, Edge]] = dict()
    edge_history: Dict[Tuple[str, str], List[TestResult]] = dict()
    action_history: List[ActionHistoryStep] = []


class GraphManager(GraphBaseAccessMixin, BaseGraphInterface):
    """
    The graph represents the internal data structure of causy. It is a simple graph with nodes and edges.
    But it supports to be handled as a directed graph, undirected graph and bidirected graph, which is important to implement different algorithms in different stages.
    It also stores the history of the actions taken on the graph.
    """

    @property
    def nodes(self) -> OrderedDict[str, Node]:
        return self.graph.nodes

    @property
    def edges(self) -> Dict[str, Dict[str, Edge]]:
        return self.graph.edges

    @property
    def _reverse_edges(self) -> Dict[str, Dict[str, Edge]]:
        return self.graph._reverse_edges

    @property
    def _deleted_edges(self) -> Dict[str, Dict[str, Edge]]:
        return self.graph._deleted_edges

    @property
    def edge_history(self) -> Dict[Tuple[str, str], List[TestResult]]:
        return self.graph.edge_history

    @property
    def action_history(self) -> List[Dict[str, List[TestResult]]]:
        return self.graph.action_history

    graph: Optional[Graph] = None

    def __init__(self, graph_class=Graph):
        self.graph = graph_class()
        self.graph.nodes = collections.OrderedDict({})
        self.graph.edges = self.__init_dict()
        self.graph._reverse_edges = self.__init_dict()
        self.graph.edge_history = self.__init_dict()
        self.graph.action_history = []

    def __init_dict(self):
        """
        Initialize a dictionary - we encapsulate this to make it easier to change the implementation of the dictionary
        As this might be necessary for multiprocessing
        :return:
        """
        return dict()

    def get_edge(self, u: Node, v: Node) -> Edge:
        """
        Get an edge between two nodes
        :param u: u node
        :param v: v node
        :return: the edge
        """
        if u.id not in self.edges:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.edges[u.id]:
            raise GraphError(f"Edge {u} -> {v} does not exist")
        return self.edges[u.id][v.id]

    def add_edge(
        self,
        u: Node,
        v: Node,
        metadata: Dict,
        edge_type: EdgeTypeInterface = UndirectedEdge(),
    ):
        """
        Add an edge to the graph
        :param edge_type: the type of the edge (e.g. undirected, directed, bidirected)
        :param u: u node
        :param v: v node
        :param metadata: metadata of the edge
        :return:
        """

        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id == v.id:
            raise GraphError("Self loops are currently not allowed")

        if u.id not in self.edges:
            self.edges[u.id] = self.__init_dict()
            self._reverse_edges[u.id] = self.__init_dict()
            self._deleted_edges[u.id] = self.__init_dict()
        if v.id not in self.edges:
            self.edges[v.id] = self.__init_dict()
            self._reverse_edges[v.id] = self.__init_dict()
            self._deleted_edges[v.id] = self.__init_dict()

        a_edge = Edge(u=u, v=v, edge_type=edge_type, metadata=metadata)
        self.edges[u.id][v.id] = a_edge
        self._reverse_edges[v.id][u.id] = a_edge

        b_edge = Edge(u=v, v=u, edge_type=edge_type, metadata=metadata)
        self.edges[v.id][u.id] = b_edge
        self._reverse_edges[u.id][v.id] = b_edge

        self.edge_history[(u.id, v.id)] = []
        self.edge_history[(v.id, u.id)] = []

    def add_directed_edge(
        self,
        u: Node,
        v: Node,
        metadata: Dict,
        edge_type: EdgeTypeInterface = DirectedEdge(),
    ):
        """
        Add a directed edge from u to v to the graph
        :param edge_type:
        :param u: u node
        :param v: v node
        :param metadata: metadata of the edge
        :return:
        """

        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id == v.id:
            raise GraphError("Self loops are currently not allowed")

        if u.id not in self.edges:
            self.edges[u.id] = self.__init_dict()
            self._deleted_edges[u.id] = self.__init_dict()
        if v.id not in self._reverse_edges:
            self._reverse_edges[v.id] = self.__init_dict()

        edge = Edge(u=u, v=v, edge_type=edge_type, metadata=metadata)

        self.edges[u.id][v.id] = edge
        self._reverse_edges[v.id][u.id] = edge

        self.edge_history[(u.id, v.id)] = []

    def add_edge_history(self, u, v, action: TestResult):
        """
        Add an action to the edge history
        :param u:
        :param v:
        :param action:
        :return:
        """
        if (u.id, v.id) not in self.edge_history:
            self.edge_history[(u.id, v.id)] = []
        self.edge_history[(u.id, v.id)].append(action)

    def remove_edge(self, u: Node, v: Node, soft_delete: bool = False):
        """
        Remove an edge from the graph (undirected)
        :param u: u node
        :param v: v node
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if soft_delete:
            if u.id in self.edges and v.id in self.edges[u.id]:
                self.edges[u.id][v.id].deleted = True
                self._reverse_edges[v.id][u.id].deleted = True

            if v.id in self.edges and u.id in self.edges[v.id]:
                self.edges[v.id][u.id].deleted = True
                self._reverse_edges[u.id][v.id].deleted = True

            return

        if u.id in self.edges and v.id in self.edges[u.id]:
            self._deleted_edges[u.id][v.id] = self.edges[u.id][v.id]
            del self.edges[u.id][v.id]
            del self._reverse_edges[u.id][v.id]

        if v.id in self.edges and u.id in self.edges[v.id]:
            self._deleted_edges[v.id][u.id] = self.edges[v.id][u.id]
            del self.edges[v.id][u.id]
            del self._reverse_edges[v.id][u.id]

    def restore_edge(self, u: Node, v: Node):
        """
        Restore a deleted edge
        :param u:
        :param v:
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id in self.edges and v.id in self.edges[u.id]:
            self.edges[u.id][v.id].deleted = False
            self._reverse_edges[v.id][u.id].deleted = False
        else:
            self.add_edge(u, v, self._deleted_edges[u.id][v.id].metadata)
            del self._deleted_edges[u.id][v.id]

        if v.id in self.edges and u.id in self.edges[v.id]:
            self.edges[v.id][u.id].deleted = False
            self._reverse_edges[u.id][v.id].deleted = False
        else:
            self.add_edge(v, u, self._deleted_edges[v.id][u.id].metadata)
            del self._deleted_edges[v.id][u.id]

    def remove_directed_edge(self, u: Node, v: Node, soft_delete: bool = False):
        """
        :param u:
        :param v:
        :param soft_delete: does not remove the edge, but marks it as deleted (useful in multithreading)
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id not in self.edges:
            return  # no edges from u
        if v.id not in self.edges[u.id]:
            return

        if soft_delete:
            self.edges[u.id][v.id].deleted = True
            self._reverse_edges[v.id][u.id].deleted = True
            return

        self._deleted_edges[u.id][v.id] = self.edges[u.id][v.id]

        del self.edges[u.id][v.id]
        del self._reverse_edges[v.id][u.id]

    def restore_directed_edge(self, u: Node, v: Node):
        """
        Restore a soft deleted edge
        :param u:
        :param v:
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")

        if u.id in self.edges and v.id in self.edges[u.id]:
            self.edges[u.id][v.id].deleted = False
            self._reverse_edges[v.id][u.id].deleted = False
        else:
            self.add_directed_edge(u, v, self._deleted_edges[u.id][v.id].metadata)
            del self._deleted_edges[u.id][v.id]

    def update_edge(
        self,
        u: Node,
        v: Node,
        metadata: Dict = None,
        edge_type: EdgeTypeInterface = None,
    ):
        """
        Update an undirected edge in the graph
        :param u: u node
        :param v: v node
        :param value:
        :param edge_type:
        :return:
        """

        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")
        if u.id not in self.edges:
            raise GraphError(f"Node {u} does not have any edges")
        if v.id not in self.edges:
            raise GraphError(f"Node {v} does not have any edges")

        if u.id not in self.edges[v.id]:
            raise GraphError(f"There is no edge from {u} to {v}")

        if v.id not in self.edges[u.id]:
            raise GraphError(f"There is no edge from {v} to {u}")

        if metadata is not None:
            obj = self.edges[u.id][v.id]
            obj.metadata = metadata
            self.edges[u.id][v.id] = obj

            obj = self.edges[v.id][u.id]
            obj.metadata = metadata
            self.edges[v.id][u.id] = obj

            obj = self._reverse_edges[u.id][v.id]
            obj.metadata = metadata
            self._reverse_edges[u.id][v.id] = obj

            obj = self._reverse_edges[v.id][u.id]
            obj.metadata = metadata
            self._reverse_edges[v.id][u.id] = obj

        if edge_type is not None:
            obj = self.edges[u.id][v.id]
            obj.edge_type = edge_type
            self.edges[u.id][v.id] = obj

            obj = self.edges[v.id][u.id]
            obj.edge_type = edge_type
            self.edges[v.id][u.id] = obj

            obj = self._reverse_edges[u.id][v.id]
            obj.edge_type = edge_type
            self._reverse_edges[u.id][v.id] = obj

            obj = self._reverse_edges[v.id][u.id]
            obj.edge_type = edge_type
            self._reverse_edges[v.id][u.id] = obj

    def update_directed_edge(
        self,
        u: Node,
        v: Node,
        metadata: Dict = None,
        edge_type: EdgeTypeInterface = None,
    ):
        """
        Update an edge in the graph
        :param u: u node
        :param v: v node
        :return:
        """
        if u.id not in self.nodes:
            raise GraphError(f"Node {u} does not exist")
        if v.id not in self.nodes:
            raise GraphError(f"Node {v} does not exist")
        if u.id not in self.edges:
            raise GraphError(f"Node {u} does not have any edges")
        if v.id not in self.edges[u.id]:
            raise GraphError(f"There is no edge from {u} to {v}")

        if metadata is not None:
            obj = self.edges[u.id][v.id]
            obj.metadata = metadata
            self.edges[u.id][v.id] = obj

            obj = self._reverse_edges[v.id][u.id]
            obj.metadata = metadata
            self._reverse_edges[v.id][u.id] = obj

        if edge_type is not None:
            obj = self.edges[u.id][v.id]
            obj.edge_type = edge_type
            self.edges[u.id][v.id] = obj

            obj = self._reverse_edges[v.id][u.id]
            obj.edge_type = edge_type
            self._reverse_edges[v.id][u.id] = obj

    def purge_soft_deleted_edges(self):
        """
        Remove all edges that are marked as soft deleted from the graph
        """
        edges_to_remove = []
        for u_id, edges in self.edges.items():
            for v_id, edge in edges.items():
                if edge.deleted is True:
                    edges_to_remove.append((self.nodes[u_id], self.nodes[v_id]))

        for u, v in edges_to_remove:
            self.remove_directed_edge(u, v)

    def add_node(
        self,
        name: str,
        values: Union[List[float], torch.Tensor],
        id_: str = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Node:
        """
        Add a node to the graph
        :param metadata: add metadata to the node
        :param name: name of the node
        :param values: values of the node
        :param id_: id_ of the node
        :param : node

        :return: created Node
        """
        if id_ is None:
            id_ = str(uuid4())

        if id_ in self.nodes:
            raise ValueError(f"Node with id {id_} already exists")

        if isinstance(values, torch.Tensor):
            tensor_values = values
        else:
            try:
                tensor_values = torch.tensor(values, dtype=torch.float64)
            except TypeError as e:
                raise ValueError(f"Currently only numeric values are supported. {e}")

        if metadata is None:
            metadata = {}

        node = Node(name=name, id=id_, values=tensor_values, metadata=metadata)

        self.nodes[id_] = node
        return node



import itertools
import logging
import random
from typing import Optional, Union, Dict

from pydantic import BaseModel

from causy.interfaces import (
    GeneratorInterface,
    BaseGraphInterface,
    GraphModelInterface,
    AS_MANY_AS_FIELDS,
)
from causy.models import ComparisonSettings
from causy.graph_utils import load_pipeline_artefact_by_definition
from causy.variables import IntegerParameter, BoolParameter

logger = logging.getLogger(__name__)


class AllCombinationsGenerator(GeneratorInterface):
    """
    Generates all combinations of nodes in the graph in accordance with the configured amount of nodes to compare.
    Yields node combinations as tuples, but only in one ordering, i.e. mathematically it returns sets.
    e.g. if your graph consists of the nodes [X, Y, Z] and you want to compare 2 nodes,
    it will yield (X, Y), (X, Z), (Y, Z). It will not additionally yield (Y, X), (Z, X), (Z, Y).
    """

    def generate(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        start = self.comparison_settings.min
        # if min (size of node sets to generate) is larger than the number of variables (represented by nodes) in our dataset, we can't create any combinations
        if start > len(graph.nodes):
            return

        # if max is AS_MANY_AS_FIELDS, we set it to number of variables + 1
        if self.comparison_settings.max == AS_MANY_AS_FIELDS:
            stop = len(graph.nodes) + 1
        else:
            stop = self.comparison_settings.max + 1

        # if start is higher than number of variables, we set it to the amount of variables
        if stop > len(graph.nodes) + 1:
            stop = len(graph.nodes) + 1

        # if stop is smaller than start, we can't create any combinations
        if stop < start:
            return

        # create all combinations
        for range_size in range(start, stop):
            for subset in itertools.combinations(graph.nodes, range_size):
                yield subset


class PairsWithEdgesInBetweenGenerator(GeneratorInterface):
    """
    Generates all pairs of nodes that have edges in between them. It does not matter if the edge is directed or not.
    However, if it is an edge which points in both/no directions, it will be iterated over them twice.
    """

    chunk_size: IntegerParameter = 100
    chunked: Optional[BoolParameter] = None

    def __init__(
        self,
        *args,
        chunk_size: Optional[IntegerParameter] = None,
        chunked: Optional[BoolParameter] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.chunked = chunked
        if chunk_size is not None:
            self.chunk_size = chunk_size

    def generate(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        edges = []

        for f_node in graph.edges.keys():
            for t_node in graph.edges[f_node].keys():
                if not graph.directed_edge_exists(f_node, t_node):
                    continue
                edges.append((f_node, t_node))

        if self.chunked:
            for i in range(0, len(edges), self.chunk_size):
                yield edges[i : i + self.chunk_size]

        for edge in edges:
            yield edge


class PairsWithNeighboursGenerator(GeneratorInterface):
    """
    Generates all combinations of pairs of nodes that are neighbours and the neighbours of the first input node.
    e.g. if your graph consists of the nodes [X, Y, Z, W, V] your output could be:
    [X, Y, neighbours(X)], [Y, X, neighbours(Y)], [X, Z, neighbours(X)], [Z, X, neighbours(Z)], ...
    (if, among others, X and Y are neighbours and X and Z are neighbours)
    """

    shuffle_combinations: BoolParameter = True
    chunked: BoolParameter = True

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

    def generate(
        self, graph: BaseGraphInterface, graph_model_instance_: GraphModelInterface
    ):
        start = self.comparison_settings.min
        # if min is longer then our dataset, we can't create any combinations
        if start > len(graph.nodes):
            return

        # if max is AS_MANY_AS_FIELDS, we set it to the length of the dataset + 1
        if self.comparison_settings.max == AS_MANY_AS_FIELDS:
            stop = len(graph.nodes) + 1
        else:
            stop = self.comparison_settings.max + 1

        # if start is longer then our dataset, we set it to the length of the dataset
        if stop > len(graph.nodes) + 1:
            stop = len(graph.nodes) + 1

        # if stop is smaller then start, we can't create any combinations
        if stop < start:
            return

        if start < 2:
            raise ValueError("PairsWithNeighboursGenerator: start must be at least 2")
        for range_size in range(start, stop):
            logger.info(f"range_size = {range_size}")
            logger.debug(f"PairsWithNeighboursGenerator: range_size={range_size}")
            checked_combinations = set()
            for node in graph.edges:
                for neighbour in graph.edges[node].keys():
                    if not graph.directed_edge_exists(node, neighbour):
                        continue

                    if (node, neighbour) in checked_combinations:
                        continue

                    checked_combinations.add((node, neighbour))
                    if range_size == 2:
                        yield [[node, neighbour]]
                        continue

                    other_neighbours = set(
                        [
                            k
                            for k, value in graph.edges[node].items()
                            if graph.directed_edge_exists(k, node)
                        ]
                    )
                    logger.info(f"other_neighbors before removal={other_neighbours}")

                    if neighbour in other_neighbours:
                        other_neighbours.remove(neighbour)
                    else:
                        logger.debug(
                            "PairsWithNeighboursGenerator: neighbour not in other_neighbours. This should not happen."
                        )
                    logger.info(f"node={node}, neighbour={neighbour}")
                    logger.info(f"other_neighbours={other_neighbours}")
                    combinations = list(
                        itertools.combinations(other_neighbours, range_size - 2)
                    )

                    if self.shuffle_combinations:
                        combinations = list(combinations)
                        import random

                        random.shuffle(combinations)

                    if self.chunked:
                        chunk = []
                        for k in combinations:
                            chunk.append([node, neighbour] + [ks for ks in k])
                        yield chunk
                    else:
                        for k in combinations:
                            yield [node, neighbour] + [ks for ks in k]


class RandomSampleGenerator(GeneratorInterface, BaseModel):
    """
    Executes another generator and returns a random sample of the results
    """

    every_nth: IntegerParameter = 100
    generator: Optional[GeneratorInterface] = None

    def __init__(
        self,
        *args,
        generator: Optional[Union[GeneratorInterface, Dict[any, any]]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if generator is not None:
            if isinstance(generator, GeneratorInterface):
                self.generator = generator
            else:
                self.generator = load_pipeline_artefact_by_definition(generator)
        else:
            raise ValueError("RandomSampleGenerator: generator must be set")

    def generate(self, graph: BaseGraphInterface, graph_model_instance_: dict):
        """
        Executes another generator and returns a random sample of the results
        :param graph:
        :param graph_model_instance_:
        :return: yields a random sample of the results
        """
        for combination in self.generator.generate(graph, graph_model_instance_):
            if random.randint(0, self.every_nth) == 0:
                yield combination



from typing import Union

from causy.graph import Node
from causy.interfaces import ExtensionInterface


class InducingPathExtension(ExtensionInterface):
    class GraphAccessMixin:
        def inducing_path_exists(
            self, u: Union[Node, str], v: Union[Node, str]
        ) -> bool:
            """
            Check if an inducing path from u to v exists.
            An inducing path from u to v is a directed reference from u to v on which all mediators are colliders.
            :param u: node u
            :param v: node v
            :return: True if an inducing path exists, False otherwise
            """

            if isinstance(u, Node):
                u = u.id
            if isinstance(v, Node):
                v = v.id

            if not self.directed_path_exists(u, v):
                return False
            for path in self.directed_paths(u, v):
                for i in range(1, len(path) - 1):
                    r, w = path[i]
                    if not self.bidirected_edge_exists(r, w):
                        # TODO: check if this is correct (@sof)
                        return True
            return False



import typer
from causy.serialization import load_json
from causy.ui.server import server
from causy.workspaces.cli import _current_workspace


def ui(result_file: str = None):
    """Start the causy UI."""
    if not result_file:
        workspace = _current_workspace()
        server_config, server_runner = server(workspace=workspace)
    else:
        result = load_json(result_file)
        server_config, server_runner = server(result=result)

    typer.launch(f"http://{server_config.host}:{server_config.port}")
    typer.echo(f"🚀 Starting server at http://{server_config.host}:{server_config.port}")
    server_runner.run()



"""
Representation of a causy workspace.

## Exemplary folder structure
- causy.yaml (contains workspace wide settings for stuff like name, author, pipeline definitions)
- any_pipeline.yaml (seperate pipeline definition referenced by causy.yaml)
- custom_pipeline_step.py (a project specific pipeline step)
- some_data_loader.py (a data loader to ship data into causy)
- tune_pipeline_1709406402/ (a single experiment)
    - causy_experiment.json (contains the full definition of everything used)
    - causy_model.json (the actual generated model)

## Examples
- the user generates a new workspace because they want to work on a new causal discovery project (causy workspace create)
- they add a custom data loader, pipeline steps and extend an existing pipeline (causy workspace data-loaders create)
- they run multiple experiments (causy workspace experiments create, causy workspace experiments execute $EXPERIMENT_NAME)
"""



from typing import Optional, List, Dict, Union

from causy.interfaces import NodeInterface
from causy.models import Result
from causy.workspaces.models import Experiment
from pydantic import BaseModel, UUID4


class NodePosition(BaseModel):
    x: Optional[float]
    y: Optional[float]


class ExperimentVersion(BaseModel):
    version: int
    name: str


class ExtendedExperiment(Experiment):
    versions: Optional[List[ExperimentVersion]] = None
    name: str = None


class PositionedNode(NodeInterface):
    position: Optional[NodePosition] = None


class ExtendedResult(Result):
    nodes: Dict[Union[UUID4, str], PositionedNode]
    version: Optional[int] = None



import copy
import enum
from types import NoneType
from typing import Any, Union, TypeVar, Generic, Optional, List, Dict

from pydantic import BaseModel, computed_field


VariableInterfaceType = TypeVar("VariableInterfaceType")

VariableType = Union[str, int, float, bool]


class VariableTypes(enum.Enum):
    String = "string"
    Integer = "integer"
    Float = "float"
    Bool = "bool"


class BaseVariable(BaseModel, Generic[VariableInterfaceType]):
    """
    Represents a single variable. It can be a string, int, float or bool. The type of the variable is determined by the
    type attribute.
    """

    def __init__(self, **data):
        super().__init__(**data)
        self.validate_value(self.value)

    name: str
    value: Union[str, int, float, bool]
    choices: Optional[List[Union[str, int, float, bool]]] = None

    def is_valid(self):
        return self.is_valid_value(self.value)

    def is_valid_value(self, value):
        try:
            self.validate_value(value)
            return True
        except ValueError:
            return False

    def validate_value(self, value):
        if not isinstance(value, self._PYTHON_TYPE):
            raise ValueError(
                f"Variable {self.name} is not valid."
                f" (should be {self.type} but is {type(value)})"
            )

        if self.choices and value not in self.choices:
            raise ValueError(
                f"Value {value} is not in the list of choices: {self.choices}"
            )

    @computed_field
    @property
    def type(self) -> Optional[str]:
        return self._TYPE


class StringVariable(
    BaseVariable[VariableInterfaceType], Generic[VariableInterfaceType]
):
    """
    Represents a single string variable.
    """

    value: str
    name: str

    _TYPE: Optional[str] = VariableTypes.String.value
    _PYTHON_TYPE: Optional[type] = str


class IntegerVariable(
    BaseVariable[VariableInterfaceType], Generic[VariableInterfaceType]
):
    """
    Represents a single int variable.
    """

    value: int
    name: str

    _TYPE: str = VariableTypes.Integer.value
    _PYTHON_TYPE: Optional[type] = int

    def validate_value(self, value):
        # check if the value is a boolean and raise an error
        # we do this because in python bool is a subclass of int
        if isinstance(value, bool):
            raise ValueError(
                f"Variable {self.name} is not valid."
                f" (should be {self.type} but is {type(value)})"
            )
        super().validate_value(value)


class FloatVariable(
    BaseVariable[VariableInterfaceType], Generic[VariableInterfaceType]
):
    """
    Represents a single float variable.
    """

    value: float
    name: str

    _TYPE: str = VariableTypes.Float.value
    _PYTHON_TYPE: Optional[type] = float


class BoolVariable(BaseVariable[VariableInterfaceType], Generic[VariableInterfaceType]):
    """
    Represents a single bool variable.
    """

    value: bool
    name: str

    _TYPE: str = VariableTypes.Bool.value
    _PYTHON_TYPE: Optional[type] = bool


class VariableReference(BaseModel, Generic[VariableInterfaceType]):
    """
    Represents a reference to a variable.
    """

    name: str

    @computed_field
    @property
    def type(self) -> str:
        return "reference"


VARIABLE_MAPPING = {
    VariableTypes.String.value: StringVariable,
    VariableTypes.Integer.value: IntegerVariable,
    VariableTypes.Float.value: FloatVariable,
    VariableTypes.Bool.value: BoolVariable,
}

BoolParameter = Union[bool, VariableReference]
IntegerParameter = Union[int, VariableReference]
FloatParameter = Union[float, VariableReference]
StringParameter = Union[str, VariableReference]
CausyParameter = Union[BoolParameter, IntegerParameter, FloatParameter, StringParameter]


def validate_variable_values(algorithm, variable_values: Dict[str, VariableType]):
    """
    Validate the variable values for the algorithm.
    :param algorithm:
    :param variable_values:
    :return:
    """
    algorithm_variables = {avar.name: avar for avar in algorithm.variables}

    for variable_name, variable_value in variable_values.items():
        if variable_name not in algorithm_variables.keys():
            raise ValueError(
                f"Variable {variable_name} not found in the algorithm variables."
            )
        algorithm_variables[variable_name].validate_value(variable_value)

    return True


def resolve_variables(
    variables: List[BaseVariable], variable_values: Dict[str, VariableType]
) -> Dict[str, VariableType]:
    """
    Resolve the variables from the list of variables and the variable values coming from the user.
    :param variables:
    :param variable_values:
    :return:
    """
    resolved_variables = {}
    for variable in variables:
        if variable.name in variable_values:
            resolved_variables[variable.name] = variable_values[variable.name]
        else:
            resolved_variables[variable.name] = variable.value

    return resolved_variables


def resolve_variable_to_object(obj: Any, variables):
    """
    Resolve the variables to the object.
    :param obj:
    :param variables:
    :return:
    """
    for attribute, value in obj.__dict__.items():
        if isinstance(value, VariableReference):
            if value.name in variables:
                obj.__dict__[attribute] = variables[value.name]
            else:
                raise ValueError(f'Variable "{value.name}" not found in the variables.')
        elif (
            hasattr(value, "__dict__")
            and not isinstance(value, NoneType)
            and not hasattr(value, "value")
        ):
            # we check for value because we don't want to resolve the variable if it's a variable object itself or a Enum
            obj.__dict__[attribute] = resolve_variable_to_object(value, variables)
    return obj


def resolve_variables_to_algorithm_for_pipeline_steps(pipeline_steps, variables):
    """
    Resolve the variables to the algorithm.
    :param pipeline_steps:
    :param variables:
    :return:
    """
    for k, pipeline_step in enumerate(pipeline_steps):
        pipeline_steps[k] = resolve_variable_to_object(pipeline_step, variables)
        # handle cases when we have sub-pipelines like in Loops
        if hasattr(pipeline_step, "pipeline_steps"):
            pipeline_step.pipeline_steps = (
                resolve_variables_to_algorithm_for_pipeline_steps(
                    pipeline_step.pipeline_steps, variables
                )
            )

    return pipeline_steps


def deserialize_variable(variable_dict: Dict[str, VariableType]) -> BaseVariable:
    """
    Deserialize the variable from the dictionary.
    :param variable_dict:
    :return:
    """
    if "type" not in variable_dict:
        raise ValueError("Variable type not found.")
    if variable_dict["type"] not in VARIABLE_MAPPING:
        raise ValueError(f"Variable type {variable_dict['type']} not found.")

    return VARIABLE_MAPPING[variable_dict["type"]](**variable_dict)


def deserialize_variable_references(element: object) -> object:
    """
    Deserialize the variable references from the pipeline step.
    :param pipeline_step:
    :return:
    """
    for attribute, value in element.__dict__.items():
        if isinstance(value, dict) and "type" in value and value["type"] == "reference":
            setattr(element, attribute, VariableReference(name=value["name"]))

        if hasattr(value, "__dict__"):
            setattr(element, attribute, deserialize_variable_references(value))

    if hasattr(element, "pipeline_steps"):
        element.pipeline_steps = [
            deserialize_variable_references(pipeline_step)
            for pipeline_step in element.pipeline_steps
        ]

    return element
